use std::error::Error;

use actix_web::{error, web, HttpResponse};
use futures::future;
use handlebars::Handlebars;
use lettre::{
    message::header::ContentType, AsyncSmtpTransport, AsyncTransport, Message, Tokio1Executor,
};
use log::debug;
use rand::{distributions::Alphanumeric, thread_rng, Rng};
use serde::Serialize;
use serde_json::json;
use terminwahl_typen::{AppointmentSlot, Nutzer, PlannedAppointment, RequestState, Teacher};

use crate::db::{
    self,
    read::{get_slot_by_id, get_teacher_by_id},
    Pool,
};

#[derive(Serialize)]
pub struct AppointmentTeacherSlot {
    pub teacher: Teacher,
    pub slot: AppointmentSlot,
}

impl AppointmentTeacherSlot {
    pub async fn get(pool: &Pool, teacher_id: i64, slot_id: i64) -> Result<Self, Box<dyn Error>> {
        Ok(Self {
            teacher: get_teacher_by_id(pool, teacher_id)
                .await
                .expect("Teacher not found"),
            slot: get_slot_by_id(pool, slot_id).await.expect("Slot not found"),
        })
    }
}

pub async fn save_appointments_json(
    pool: web::Data<Pool>,
    mailer: web::Data<AsyncSmtpTransport<Tokio1Executor>>,
    handlebars: web::Data<Handlebars<'_>>,
    input: web::Json<(Vec<PlannedAppointment>, Nutzer)>,
) -> Result<HttpResponse, error::Error> {
    debug!("Extracting data");
    let (appointments, nutzer) = input.into_inner();
    debug!("Saving user");
    let nutzer_id = db::write::save_nutzer(&pool, &nutzer)
        .await
        .map_err(error::ErrorInternalServerError)?;
    debug!("Saving appointments");
    let validation_key: String = thread_rng()
        .sample_iter(&Alphanumeric)
        .take(30)
        .map(char::from)
        .collect();
    db::write::save_appointments(&pool, &appointments, nutzer_id, &validation_key)
        .await
        .map_err(error::ErrorInternalServerError)?;

    debug!("Loading the appointments");
    let full_appointments = future::try_join_all(appointments.into_iter().map(
        |PlannedAppointment {
             teacher_id,
             slot_id,
         }| { AppointmentTeacherSlot::get(&pool, teacher_id, slot_id) },
    ))
    .await
    .expect("Failed to get full list of appointments");
    debug!("Send the email");
    let mail_result = send_confirmation_request(
        &full_appointments,
        &nutzer,
        &validation_key,
        handlebars.as_ref(),
        &mailer,
    )
    .await;

    Ok(HttpResponse::Ok().json(mail_result))
}

pub async fn send_confirmation_request<'a>(
    appointments: &Vec<AppointmentTeacherSlot>,
    nutzer: &Nutzer,
    validation_key: &str,
    handlebars: &Handlebars<'a>,
    mailer: &AsyncSmtpTransport<Tokio1Executor>,
) -> RequestState {
    let data = json! {
    {
        "appointments": appointments,
        "nutzer": nutzer,
        "validation_key": validation_key
    }};
    debug!("{:?}", handlebars.get_templates());
    match handlebars.render("email_confirm", &data) {
        Ok(email_text) => {
            let email = match Message::builder()
                .from(
                    "Franz Dietrich <franz.dietrich@uhlandshoehe.de>"
                        .parse()
                        .expect("Should not fail"),
                )
                .to(
                    match format!("{} <{}>", nutzer.name, nutzer.email).parse() {
                        Ok(v) => v,
                        Err(e) => {
                            debug!("Error while sending email {e}");
                            return RequestState::Error;
                        }
                    },
                )
                .subject("Elternsprechtag: Bestätigen Sie Ihre Termine")
                .header(ContentType::TEXT_PLAIN)
                .body(
                    match lettre::message::Body::new_with_encoding(
                        email_text,
                        lettre::message::header::ContentTransferEncoding::Base64,
                    ) {
                        Ok(body) => body,
                        Err(e) => {
                            debug!("Error while sending email {e:?}");
                            return RequestState::Error;
                        }
                    },
                ) {
                Ok(message) => message,
                Err(e) => {
                    debug!("Error while sending email {e}");
                    return RequestState::Error;
                }
            };

            // Send the email
            match mailer.send(email).await {
                Ok(_) => RequestState::Success,
                Err(e) => {
                    debug!("Failed to send: {e}");
                    RequestState::Error
                }
            }
        }
        Err(e) => {
            debug!("Failed to send email: Template did not render {e}");
            RequestState::Error
        }
    }
}
